(* *********************************************************************)
(*                                                                     *)
(*              The Compcert verified compiler                         *)
(*                                                                     *)
(*          Xavier Leroy, INRIA Paris-Rocquencourt                     *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique.  All rights reserved.  This file is distributed       *)
(*  under the terms of the INRIA Non-Commercial License Agreement.     *)
(*                                                                     *)
(* *********************************************************************)

(** The Mach intermediate language: abstract syntax.

  Mach is the last intermediate language before generation of assembly
  code.  
*)

Require Import Coqlib.
Require Import Maps.
Require Import AST.
Require Import Integers.
Require Import Values.
Require Import Memory.
Require Import Globalenvs.
Require Import Events.
Require Import Smallstep.
Require Import Op.
Require Import Locations.
Require Import Conventions.
Require Stacklayout.

(** * Abstract syntax *)

(** Like Linear, the Mach language is organized as lists of instructions
  operating over machine registers, with default fall-through behaviour
  and explicit labels and branch instructions.  

  The main difference with Linear lies in the instructions used to
  access the activation record.  Mach has three such instructions:
  [Mgetstack] and [Msetstack] to read and write within the activation
  record for the current function, at a given word offset and with a
  given type; and [Mgetparam], to read within the activation record of
  the caller. 

  These instructions implement a more concrete view of the activation
  record than the the [Lgetstack] and [Lsetstack] instructions of
  Linear: actual offsets are used instead of abstract stack slots, and the
  distinction between the caller's frame and the callee's frame is
  made explicit. *)

Definition label := positive.

Inductive instruction: Type :=
  | Mgetstack: int -> typ -> mreg -> instruction
  | Msetstack: mreg -> int -> typ -> instruction
  | Mgetparam: int -> typ -> mreg -> instruction
  | Mop: operation -> list mreg -> mreg -> instruction
  | Mload: memory_chunk -> addressing -> list mreg -> mreg -> instruction
  | Mstore: memory_chunk -> addressing -> list mreg -> mreg -> instruction
  | Mcall: signature -> mreg + ident -> instruction
  | Mtailcall: signature -> mreg + ident -> instruction
  | Mbuiltin: external_function -> list mreg -> mreg -> instruction
  | Mannot: external_function -> list annot_param -> instruction
  | Mlabel: label -> instruction
  | Mgoto: label -> instruction
  | Mcond: condition -> list mreg -> label -> instruction
  | Mjumptable: mreg -> list label -> instruction
  | Mreturn: instruction

with annot_param: Type :=
  | APreg: mreg -> annot_param
  | APstack: memory_chunk -> Z -> annot_param.

Definition code := list instruction.

Record function: Type := mkfunction
  { fn_id: ident;
    fn_sig: signature;
    fn_code: code;
    fn_stacksize: Z(*;
    fn_link_ofs: int ;
    fn_retaddr_ofs: int *) }.

Definition fundef := AST.fundef function.

Definition program := AST.program fundef unit.

Definition funsig (fd: fundef) :=
  match fd with
  | Internal f => fn_sig f
  | External ef => ef_sig ef
  end.

Definition genv := Genv.t fundef unit.

(** * Operational semantics *)

(** The semantics for Mach is close to that of [Linear]: they differ only
  on the interpretation of stack slot accesses.  In Mach, these
  accesses are interpreted as memory accesses relative to the
  stack pointer.  More precisely:
- [Mgetstack ofs ty r] is a memory load at offset [ofs * 4] relative
  to the stack pointer.
- [Msetstack r ofs ty] is a memory store at offset [ofs * 4] relative
  to the stack pointer.
- [Mgetparam ofs ty r] is a memory load at offset [ofs * 4]
  relative to the pointer found at offset 0 from the stack pointer.
  The semantics maintain a linked structure of activation records,
  with the current record containing a pointer to the record of the
  caller function at offset 0.

In addition to this linking of activation records, the
semantics also make provisions for storing a back link at offset
[f.(fn_link_ofs)] from the stack pointer, and a return address at
offset [f.(fn_retaddr_ofs)].  The latter stack location will be used
by the Asm code generated by [Asmgen] to save the return address into
the caller at the beginning of a function, then restore it and jump to
it at the end of a function.  The Mach concrete semantics does not
attach any particular meaning to the pointer stored in this reserved
location, but makes sure that it is preserved during execution of a
function.  The [return_address_offset] parameter is used to guess the
value of the return address that the Asm code generated later will
store in the reserved location.
*)

Definition chunk_of_type (ty: typ) :=
  match ty with Tint => Mint32 | Tfloat => Mfloat64al32 end.

Definition load_stack (m: mem) (sp: val) (ty: typ) (ofs: int) :=
  Mem.loadv (chunk_of_type ty) m (Val.add sp (Vint ofs)).

Definition store_stack (m: mem) (sp: val) (ty: typ) (ofs: int) (v: val) :=
  Mem.storev (chunk_of_type ty) m (Val.add sp (Vint ofs)) v.

Module RegEq.
  Definition t := mreg.
  Definition eq := mreg_eq.
End RegEq.

Module Regmap := EMap(RegEq).

Definition regset := Regmap.t val.

Notation "a ## b" := (List.map a b) (at level 1).
Notation "a # b <- c" := (Regmap.set b c a) (at level 1, b at next level).

Fixpoint undef_regs (rl: list mreg) (rs: regset) {struct rl} : regset :=
  match rl with
  | nil => rs
  | r1 :: rl' => undef_regs rl' (Regmap.set r1 Vundef rs)
  end.

Lemma undef_regs_other:
  forall r rl rs, ~In r rl -> undef_regs rl rs r = rs r.
Proof.
  induction rl; simpl; intros. auto. rewrite IHrl. apply Regmap.gso. intuition. intuition.
Qed.

Lemma undef_regs_same:
  forall r rl rs, In r rl \/ rs r = Vundef -> undef_regs rl rs r = Vundef.
Proof.
  induction rl; simpl; intros. tauto.
  destruct H. destruct H. apply IHrl. right. subst; apply Regmap.gss.
  auto.
  apply IHrl. right. unfold Regmap.set. destruct (RegEq.eq r a); auto. 
Qed.

Definition undef_temps (rs: regset) := 
  undef_regs temporary_regs rs.

Definition undef_move (rs: regset) := 
  undef_regs destroyed_at_move_regs rs.

Definition undef_op (op: operation) (rs: regset) :=
  match op with
  | Omove => undef_move rs
  | _ => undef_temps rs
  end.

Definition undef_setstack (rs: regset) := undef_move rs.

Definition is_label (lbl: label) (instr: instruction) : bool :=
  match instr with
  | Mlabel lbl' => if peq lbl lbl' then true else false
  | _ => false
  end.

Lemma is_label_correct:
  forall lbl instr,
  if is_label lbl instr then instr = Mlabel lbl else instr <> Mlabel lbl.
Proof.
  intros.  destruct instr; simpl; try discriminate.
  case (peq lbl l); intro; congruence.
Qed.

Fixpoint find_label (lbl: label) (c: code) {struct c} : option code :=
  match c with
  | nil => None
  | i1 :: il => if is_label lbl i1 then Some il else find_label lbl il
  end.

Lemma find_label_incl:
  forall lbl c c', find_label lbl c = Some c' -> incl c' c.
Proof.
  induction c; simpl; intros. discriminate.
  destruct (is_label lbl a). inv H. auto with coqlib. eauto with coqlib. 
Qed.

Section RELSEM.

(* Variable return_address_offset: function -> code -> int -> Prop. *)

Variable ge: genv.

Definition find_function_ptr
        (ge: genv) (ros: mreg + ident) (rs: regset) : option block :=
  match ros with
  | inl r =>
      match rs r with
      | Vptr b ofs =>
        if Int.eq ofs Int.zero then
          match Genv.find_funct_ptr ge b with
              | None => None
              | Some _ => Some b
          end
        else None
      | _ => None
      end
  | inr symb =>
      Genv.find_symbol ge symb
  end.

(** Extract the values of the arguments to an external call. *)

Inductive extcall_arg: regset -> mem -> val -> loc -> val -> Prop :=
  | extcall_arg_reg: forall rs m sp r,
      extcall_arg rs m sp (R r) (rs r)
  | extcall_arg_stack: forall rs m sp ofs ty v,
      load_stack m sp ty (Int.repr (Stacklayout.fe_ofs_arg + 4 * ofs)) = Some v ->
      extcall_arg rs m sp (S (Outgoing ofs ty)) v.

Lemma extcall_arg_determ: forall r m v l v1,
                            extcall_arg r m v l v1 ->
                            forall v2, 
                              extcall_arg r m v l v2 ->
                              v1 = v2.
Proof.
  inversion 1; subst; inversion 1; subst; congruence.
Qed.

Definition extcall_arguments
    (rs: regset) (m: mem) (sp: val) (sg: signature) (args: list val) : Prop :=
  list_forall2 (extcall_arg rs m sp) (loc_arguments sg) args.

Lemma extcall_arguments_determ':
  forall rs m sp l args1,
    list_forall2 (extcall_arg rs m sp) l args1 ->
    forall args2,
    list_forall2 (extcall_arg rs m sp) l args2 ->
    args1 = args2.
Proof.
  induction 1; inversion 1; subst.
   reflexivity.
  exploit extcall_arg_determ. eassumption. eexact H.
  exploit IHlist_forall2. eassumption.
  congruence.
Qed.

Lemma extcall_arguments_determ:
  forall rs m sp sg args1,
    extcall_arguments rs m sp sg args1 ->
    forall args2, extcall_arguments rs m sp sg args2 ->
                  args1 = args2.
Proof.
  intros; eauto using extcall_arguments_determ'.
Qed. 

(** Extract the values of the arguments to an annotation. *)

Inductive annot_arg: regset -> mem -> val -> annot_param -> val -> Prop :=
  | annot_arg_reg: forall rs m sp r,
      annot_arg rs m sp (APreg r) (rs r)
  | annot_arg_stack: forall rs m stk base chunk ofs v,
      Mem.load chunk m stk (Int.unsigned base + ofs) = Some v ->
      annot_arg rs m (Vptr stk base) (APstack chunk ofs) v.

Lemma annot_arg_determ:
  forall r m v a v1,
    annot_arg r m v a v1 ->
    forall v2, annot_arg r m v a v2 ->
               v1 = v2.
Proof.
  inversion 1; subst; inversion 1; subst; congruence.
Qed.

Definition annot_arguments
   (rs: regset) (m: mem) (sp: val) (params: list annot_param) (args: list val) : Prop :=
  list_forall2 (annot_arg rs m sp) params args.

Lemma annot_arguments_determ: forall rs m sp params args1,
                                annot_arguments rs m sp params args1 ->
                                forall args2,
                                  annot_arguments rs m sp params args2 ->
                                  args1 = args2.
Proof.
  induction 1; inversion 1; subst.
   reflexivity.
   exploit annot_arg_determ.
   eassumption.
   eexact H.
   exploit IHlist_forall2.
   eassumption.
   congruence.
Qed.

(** Mach execution states. *)

Inductive stackframe: Type :=
  | Stackframe:
      forall (f: block)       (**r pointer to calling function *)
             (sp: block)        (**r stack pointer in calling function *)
(*             (retaddr: val)   (**r Asm return address in calling function *) *)
             (c: code),       (**r program point in calling function *)
      stackframe.

Inductive state: Type :=
  | State:
      forall (stack: list stackframe)  (**r call stack *)
             (f: block)                (**r pointer to current function *)
             (sp: block)                 (**r stack pointer *)
             (c: code)                 (**r current program point *)
             (rs: regset)              (**r register state *)
             (m: mem),                 (**r memory state *)
      state
  | Callstate:
      forall
             (stack: list stackframe)  (**r call stack *)
             (f: block)                (**r pointer to function to call *)
             (rs: regset)              (**r register state *)
             (m: mem),                 (**r memory state *)
      state
  | Returnstate:
      forall (stack: list stackframe)  (**r call stack *)
             (rs: regset)              (**r register state *)
             (m: mem),                 (**r memory state *)
      state.

Definition parent_sp (s: list stackframe) : val :=
  match s with
  | nil => Vzero (* Vptr Mem.nullptr Int.zero *)
  | Stackframe f sp (* ra *) c :: s' => Vptr sp Int.zero
  end.

(*
Definition parent_ra (s: list stackframe) : val :=
  match s with
  | nil => Vzero
  | Stackframe f sp (* ra *) c :: s' => ra
  end.
*)

Inductive step: state -> trace -> state -> Prop :=
  | exec_Mlabel:
      forall s f sp lbl c rs m,
      step (State s f sp (Mlabel lbl :: c) rs m)
        E0 (State s f sp c rs m)
  | exec_Mgetstack:
      forall s f sp ofs ty dst c rs m v,
      load_stack m (Vptr sp Int.zero) ty ofs = Some v ->
      step (State s f sp (Mgetstack ofs ty dst :: c) rs m)
        E0 (State s f sp c (rs#dst <- v) m)
  | exec_Msetstack:
      forall s f sp src ofs ty c rs m m',
      store_stack m (Vptr sp Int.zero) ty ofs (rs src) = Some m' ->
      step (State s f sp (Msetstack src ofs ty :: c) rs m)
        E0 (State s f sp c (undef_setstack rs) m')
  | exec_Mgetparam:
      forall s fb f sp ofs ty dst c rs m v,
      Genv.find_funct_ptr ge fb = Some (Internal f) ->
      (* load_stack m (Vptr sp Int.zero) Tint f.(fn_link_ofs) = Some (parent_sp s) *) True -> (* It is now useless to store the back pointer *)
      load_stack m (parent_sp s) ty ofs = Some v ->
      step (State s fb sp (Mgetparam ofs ty dst :: c) rs m)
        E0 (State s fb sp c (rs # IT1 <- Vundef # dst <- v) m)
  | exec_Mop:
      forall s f sp op args res c rs m v,
      eval_operation ge (Vptr sp Int.zero) op rs##args m = Some v ->
      step (State s f sp (Mop op args res :: c) rs m)
        E0 (State s f sp c ((undef_op op rs)#res <- v) m)
  | exec_Mload:
      forall s f sp chunk addr args dst c rs m a v,
      eval_addressing ge (Vptr sp Int.zero) addr rs##args = Some a ->
      Mem.loadv chunk m a = Some v ->
      step (State s f sp (Mload chunk addr args dst :: c) rs m)
        E0 (State s f sp c ((undef_temps rs)#dst <- v) m)
  | exec_Mstore:
      forall s f sp chunk addr args src c rs m m' a,
      eval_addressing ge (Vptr sp Int.zero) addr rs##args = Some a ->
      Mem.storev chunk m a (rs src) = Some m' ->
      step (State s f sp (Mstore chunk addr args src :: c) rs m)
        E0 (State s f sp c (undef_temps rs) m')
  | exec_Mcall:
      forall s fb sp sig ros c rs m f f' (* ra *),
      find_function_ptr ge ros rs = Some f' ->
      Genv.find_funct_ptr ge fb = Some (Internal f) ->
      True (* return_address_offset f c ra *) ->
      step (State s fb sp (Mcall sig ros :: c) rs m)
        E0 (Callstate (Stackframe fb sp (* Vptr fb ra *) c :: s)
                       f' rs m)
  | exec_Mtailcall:
      forall s fb stk sig ros c rs m f f' m',
      find_function_ptr ge ros rs = Some f' ->
      Genv.find_funct_ptr ge fb = Some (Internal f) ->
      (* load_stack m (Vptr stk Int.zero) Tint f.(fn_link_ofs) = Some (parent_sp s) *) True ->
      True (* load_stack m (Vptr stk soff) Tint f.(fn_retaddr_ofs) = Some (parent_ra s) *) ->
      Mem.free m stk 0 f.(fn_stacksize) = Some m' ->
      forall FID (HFID: FID = f.(fn_id)),
      step (State s fb stk (Mtailcall sig ros :: c) rs m)
        (Event_return FID :: E0) (Callstate s f' rs m')
  | exec_Mbuiltin:
      forall s f sp rs m ef args res b t v m',
      external_call ef ge rs##args m t v m' ->
      step (State s f sp (Mbuiltin ef args res :: b) rs m)
         t (State s f sp b ((undef_temps rs)#res <- v) m')
  | exec_Mannot:
      forall s f sp rs m ef args b vargs t v m',
      annot_arguments rs m (Vptr sp Int.zero) args vargs ->
      external_call ef ge vargs m t v m' ->
      step (State s f sp (Mannot ef args :: b) rs m)
         t (State s f sp b rs m')
  | exec_Mgoto:
      forall s fb f sp lbl c rs m c',
      Genv.find_funct_ptr ge fb = Some (Internal f) ->
      find_label lbl f.(fn_code) = Some c' ->
      step (State s fb sp (Mgoto lbl :: c) rs m)
        E0 (State s fb sp c' rs m)
  | exec_Mcond_true:
      forall s fb f sp cond args lbl c rs m c',
      eval_condition cond rs##args m = Some true ->
      Genv.find_funct_ptr ge fb = Some (Internal f) ->
      find_label lbl f.(fn_code) = Some c' ->
      step (State s fb sp (Mcond cond args lbl :: c) rs m)
        E0 (State s fb sp c' (undef_temps rs) m)
  | exec_Mcond_false:
      forall s f sp cond args lbl c rs m,
      eval_condition cond rs##args m = Some false ->
      step (State s f sp (Mcond cond args lbl :: c) rs m)
        E0 (State s f sp c (undef_temps rs) m)
  | exec_Mjumptable:
      forall s fb f sp arg tbl c rs m n lbl c',
      rs arg = Vint n ->
      list_nth_z tbl (Int.unsigned n) = Some lbl ->
      Genv.find_funct_ptr ge fb = Some (Internal f) ->
      find_label lbl f.(fn_code) = Some c' ->
      step (State s fb sp (Mjumptable arg tbl :: c) rs m)
        E0 (State s fb sp c' (undef_temps rs) m)
  | exec_Mreturn:
      forall s fb stk c rs m f m',
      Genv.find_funct_ptr ge fb = Some (Internal f) ->
      (* load_stack m (Vptr stk Int.zero) Tint f.(fn_link_ofs) = Some (parent_sp s) *) True ->
      True (* load_stack m (Vptr stk soff) Tint f.(fn_retaddr_ofs) = Some (parent_ra s) *) ->
      Mem.free m stk 0 f.(fn_stacksize) = Some m' ->
      forall FID (HFID: FID = f.(fn_id)),
      step (State s fb stk (Mreturn :: c) rs m)
        (Event_return FID :: E0) (Returnstate s rs m')
  | exec_function_internal:
      forall s fb rs m f m1 stk,
      Genv.find_funct_ptr ge fb = Some (Internal f) ->
      Mem.alloc m 0 f.(fn_stacksize) = (m1, stk) ->
      let sp := Vptr stk Int.zero in
      True (* store_stack m1 sp Tint f.(fn_link_ofs) (parent_sp s) = Some m2 *) ->
      True (* store_stack m2 sp Tint f.(fn_retaddr_ofs) (parent_ra s) = Some m3 *) ->
      forall FID (HFID: FID = f.(fn_id)),
      step (Callstate s fb rs m)
        (Event_call FID :: E0) (State s fb stk f.(fn_code) (undef_temps rs) m1)
  | exec_function_external:
      forall s fb rs m t rs' ef args res m',
      Genv.find_funct_ptr ge fb = Some (External ef) ->
      external_call ef ge args m t res m' ->
      extcall_arguments rs m (parent_sp s) (ef_sig ef) args ->
      rs' = (rs#(loc_result (ef_sig ef)) <- res) ->
      step (Callstate s fb rs m)
         t (Returnstate s rs' m')
  | exec_return:
      forall s f sp (* ra *) c rs m,
      step (Returnstate (Stackframe f sp (* ra *) c :: s) rs m)
        E0 (State s f sp c rs m).

End RELSEM.

Inductive initial_state (p: program): state -> Prop :=
  | initial_state_intro: forall fb m0,
      let ge := Genv.globalenv p in
      Genv.init_mem p = Some m0 ->
      Genv.find_symbol ge p.(prog_main) = Some fb ->
      initial_state p (Callstate nil fb (Regmap.init Vundef) m0).

Inductive final_state: state -> int -> Prop :=
  | final_state_intro: forall rs m r,
      rs (loc_result (mksignature nil (Some Tint))) = Vint r ->
      final_state (Returnstate nil rs m) r.

Definition semantics (p: program) :=
  Semantics (step) (initial_state p) final_state (Genv.globalenv p).

Lemma semantics_receptive : forall p, receptive (semantics p).
Proof.
  constructor.
  inversion 1; subst; try (inversion 1; subst; eauto; fail).
  intro. exploit external_call_receptive. eassumption. eassumption.
  destruct 1. destruct H2. esplit. econstructor. eassumption.
  intro. exploit external_call_receptive. eassumption. eassumption.
  destruct 1. destruct H3. esplit. econstructor. eassumption. eassumption.
  intro. exploit external_call_receptive. eassumption. eassumption.
  destruct 1. destruct H4. esplit. econstructor. eassumption. eassumption. assumption. reflexivity.

  intro. inversion 1; subst; simpl; eauto using external_call_trace_length.
Qed.

Lemma semantics_determinate : forall p, determinate (semantics p).
Proof.
  constructor.
  inversion 1; subst; inversion 1; subst; simpl; try (exfalso; congruence);
  (let t := split; [constructor | intros; congruence] in try t);
  (let t := split; [| congruence] in try t).
  replace f0 with f in * by congruence. constructor.
  exploit external_call_determ. eexact H0. eexact H13. destruct 1. split; auto. intro. destruct H3. assumption. congruence.
  exploit annot_arguments_determ. eassumption. eexact H0. intro; subst.
  exploit external_call_determ. eexact H1. eexact H14. destruct 1. split; auto. intro. destruct H4. assumption. congruence.
  replace f0 with f in * by congruence. constructor.
  replace f0 with f in * by congruence. constructor.
  replace ef0 with ef in * by congruence.
  exploit extcall_arguments_determ. eexact H2. eexact H12. intro; subst.
  exploit external_call_determ. eexact H1. eexact H11. destruct 1. split; auto. intro. destruct H5. assumption. congruence.

  intro. inversion 1; subst; simpl; eauto using external_call_trace_length.

  inversion 1; subst; inversion 1; subst. unfold ge0, ge in *. congruence.

  inversion 1; subst. intro. intros. intro.
  inversion H1.

  inversion 1; subst; inversion 1 ; subst; congruence.
Qed.
