(* *********************************************************************)
(*                                                                     *)
(*              The Compcert verified compiler                         *)
(*                                                                     *)
(*          Xavier Leroy, INRIA Paris-Rocquencourt                     *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique.  All rights reserved.  This file is distributed       *)
(*  under the terms of the INRIA Non-Commercial License Agreement.     *)
(*                                                                     *)
(* *********************************************************************)

(** Corollaries of the main semantic preservation theorem. *)

Require Import Classical.
Require Import Coqlib.
Require Import AST.
Require Import Integers.
Require Import Values.
Require Import Events.
Require Import Globalenvs.
Require Import Smallstep.
Require Import Behaviors.
Require Import Csyntax.
Require Import Csem.
Require Import Cstrategy.
Require Import Clight.
Require Import Cminor.
Require Import RTL.
Require Import Asm.
Require Import Compiler.
Require Import Errors.
Require Import Prune.
Require Import Memdata.

Section WITHBOUNDS.
Context
  (bound : Integers.Int.int)
  (Hbound: (Stacklayout.strong_align
       | Integers.Int.unsigned bound + size_chunk Mint32))
  (external_event_needs: Events.event -> Z)
  (p: Csyntax.program) (tp: Asm.program)
  (H: transf_c_program p = OK tp)
.

(** * Preservation of whole-program behaviors *)

(** From the simulation diagrams proved in file [Compiler]. it follows that
  whole-program observable behaviors are preserved in the following sense.
  First, every behavior of the generated assembly code is matched by
  a behavior of the source C code. *)

Section CSTRATEGY.

(** If we consider the C evaluation strategy implemented by the compiler,
  we get stronger preservation results. *)

Lemma prune_atomic_behaves_intro:
  forall beh, program_behaves (prune_semantics (Cstrategy.semantics p)) beh ->
              program_behaves (prune_semantics (atomic (Cstrategy.semantics p))) beh.
Proof.
  intros.
  destruct (prune_program_behaves _ _ H0).
  destruct H1.
  apply atomic_behaviors in H2.
  eapply program_behaves_prune.
  eassumption.
  assumption.
  intros. eapply ssr_well_behaved. apply Cstrategy.semantics_strongly_receptive.
Qed.

Lemma prune_atomic_behaves_elim:  
  forall beh, program_behaves (prune_semantics (atomic (Cstrategy.semantics p))) beh ->
              program_behaves (prune_semantics (Cstrategy.semantics p)) beh.
Proof.
  intros.
  destruct (prune_program_behaves _ _ H0).
  destruct H1.
  apply atomic_behaviors in H2.
  eapply program_behaves_prune.
  eassumption.
  assumption.
  intros. eapply ssr_well_behaved. apply Cstrategy.semantics_strongly_receptive.
Qed.  

Variables
  (NOT_STUCK: not_stuck (prune_semantics (atomic (Cstrategy.semantics p))))
  (NO_OVERFLOW: no_overflow_with_mach bound external_event_needs (fun p' => atomic (Cstrategy.semantics p')) transf_c_to_mach p)
.

Theorem transf_cstrategy_program_preservation:
  (forall beh, program_behaves (prune_semantics (Cstrategy.semantics p)) beh <->
               program_behaves (Asm.semantics bound tp) beh).
Proof.
  split.
  intros.
  eapply forward_simulation_same_safe_behavior.
  eapply transf_cstrategy_program_correct.
  assumption.
  eassumption.
  eassumption.
  assumption.
  apply prune_atomic_behaves_intro.
  assumption.
  apply NOT_STUCK.
  apply prune_atomic_behaves_intro.
  assumption.
 intro.
 apply prune_atomic_behaves_elim.
 eapply backward_simulation_same_safe_behavior.
 eapply transf_cstrategy_program_correct.
 eassumption.
 assumption.
 eassumption.
 eassumption.
 assumption.
 assumption.
Qed.

End CSTRATEGY.

Variables
  (NOT_STUCK: not_stuck (prune_semantics  (Csem.semantics p)))
  (NO_OVERFLOW: no_overflow_with_mach bound external_event_needs (Csem.semantics) transf_c_to_mach p)
.

Theorem transf_c_program_preservation:
  forall beh,
  program_behaves (Asm.semantics bound tp) beh ->
  program_behaves (prune_semantics (Csem.semantics p)) beh.
Proof.
  intros. eapply backward_simulation_same_safe_behavior; eauto.
  eapply transf_c_program_correct; eauto.
Qed.

(** * Satisfaction of specifications *)

(** The second additional results shows that if all executions
  of the source C program satisfies a given specification
  (a predicate on the observable behavior of the program),
  then all executions of the produced Asm program satisfy
  this specification as well.  

  We first show this result for specifications that are stable
  under the [behavior_improves] relation. *) 

Section SPECS_PRESERVED.

Variable spec: program_behavior -> Prop.

Theorem transf_c_program_preserves_spec:
  (forall beh, program_behaves (prune_semantics (Csem.semantics p)) beh -> spec beh) ->
  (forall beh, program_behaves (Asm.semantics bound tp) beh -> spec beh).
Proof.
  intros; eauto using transf_c_program_preservation.
Qed.

End SPECS_PRESERVED.

End WITHBOUNDS.
